library(tidyverse)
library(ggpubr)
n.rep <- 100 # number of simulated data sets
cores <- 100 # number of cores in parallel
n.strata <- 1 # number of strata
mean.size.strata <- 10000 # average size of strata
# size.strata <- rpois(n.strata, mean.size.strata) # can be user-specified
size.strata <- rep(mean.size.strata, n.strata) # user-specified
rho <- 0.6 # correlation of AR(1) covar structure
p <- 2 # number of covariates
Z.char <- paste0("z",1:p) # covariate names
k.pd <- 7 # number of knots in B-spline 6
k.cal <- 7
degree.pd <- 3 # 2
degree.cal <- 3
end.pd <- 3 # end-of-followup time
end.cal <- 5
set.seed(1234)
level <- 0.95
quant.upper <- qnorm((1+level)/2)
strata <- rep(letters[1:n.strata], size.strata) # strata IDs
times.pd <- seq(0, end.pd, end.pd/100)
times.cal <- seq(0, end.cal, end.cal/100)
beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
for (i in 1:length(times.cal)) {
  for (j in 1:length(times.pd)) {
    beta.true[i, j] <- exp(-0.5*times.cal[i])*sin(3*pi*times.pd[j]/4)
  }
}
D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
mu1 <- 0.005; mu2 <- 0.002
cl <- parallel::makeCluster(cores)
doParallel::registerDoParallel(cl)
`%dorng%` <- doRNG::`%dorng%`
`%dopar%` <- foreach::`%dopar%`
system.time(raw <-
              foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                               #.options.RNG=1234,
                               .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo"),
                               #.combine=rbind,
                               .noexport=c("surtiver_fixtra_bresties_fit",
                                           "surtiver_fixtra_bresties_fit_diffpen")) 
            %dopar% {
              Rcpp::sourceCpp("surtiver_tps.cpp")
              gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
              Z <- mvtnorm::rmvnorm(sum(size.strata),
                                    sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
              ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
              U <- runif(sum(size.strata)) # survivor function F(t)
              integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                return(c(0.5*exp(gamma+z[1]+exp(-0.5*t.cal)*sin(3*pi*t/4)*(t<=end.pd)*z[2])))
              }
              time.failure <- sapply(1:sum(size.strata), function(i) {
                res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                  integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                    log(1 - U[i]), c(0, end.pd+1)), silent=T)
                return(ifelse(class(res)=="try-error", end.pd, res$root))
              })
              time.censoring <- runif(sum(size.strata), 0, end.pd)
              data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                 time.cal=ts.cal,
                                 event=1*(time.failure<=time.censoring),
                                 strata, Z, stringsAsFactors=F)
              names(data)[-(1:4)] <- Z.char
              data <- data[order(data$strata, data$time.pd, -data$event),]
              knots.pd <- quantile(data$time.pd[data$event==1],
                                   probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
              uniqt.pd.str <- unique(data$time.pd[data$event==1])
              bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                      knots=knots.pd, degree=degree.pd,
                                      Boundary.knots=c(0, end.pd))
              knots.cal <- quantile(data$time.cal[data$event==1],
                                    probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
              bases.cal <- splines::bs(data$time.cal, intercept=T,
                                       knots=knots.cal, degree=degree.cal,
                                       Boundary.knots=c(0, end.cal))
              count.strata <- sapply(split(data$strata, data$strata), length)
              theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
              fit.pen <-
                surtiver_fixtra_bresties_fit_diffpen(
                  data$event, data$time.pd, count.strata, 
                  as.matrix(data[,Z.char[2]]),
                  as.matrix(bases.cal), as.matrix(bases.pd), 
                  theta.init, as.matrix(data[,Z.char[1]]), 
                  as.numeric(0), D1, D2, mu1, mu2)
              
              B.cal <- as.matrix(splines::bs(times.cal, degree=degree.cal,
                                             intercept=T, knots=knots.cal,
                                             Boundary.knots=c(0, end.cal)))
              B.pd <- splines::bs(times.pd, degree=degree.pd, intercept=T,
                                  knots=knots.pd, Boundary.knots=c(0, end.pd))
              info.pen <- fit.pen$info
              invinfo.pen <- solve(info.pen)
              A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                             mu1*diag(length(mu1)))
              B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                             mu2*diag(length(mu2)))
              P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
              theta.est <- fit.pen$ctrl.pts[1,] +
                invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                P%*%fit.pen$ctrl.pts[1,]
              beta.est <- B.cal%*%matrix(theta.est, byrow=T,
                                         ncol=k.pd)%*%t(B.pd)
              info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] <-
                info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] - P
              var.tv <- (invinfo.pen%*%info.pen%*%invinfo.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
              B.cal.expand <- B.cal[rep(1:nrow(B.cal), each=nrow(B.pd)),]
              B.pd.expand <- kronecker(matrix(1, nrow(B.cal), 1), B.pd)
              mat.dev <- matrix(sapply(1:(nrow(B.cal)*nrow(B.pd)),
                                       function(i) {
                                         tmp <- kronecker(B.cal.expand[i,], B.pd.expand[i,])
                                         sqrt(t(tmp)%*%var.tv%*%tmp)
                                       })*quant.upper, nrow=length(times.cal), byrow=T)
              1*(beta.est - mat.dev <= beta.true & beta.est + mat.dev >= beta.true)
            })
save(raw, file="list_of_CP_pen_matrices_n=10000.RData")
parallel::stopCluster(cl)
CP <- Reduce("+", raw)/n.rep
mytheme <- theme(plot.title=element_text(hjust=0),
                 axis.title=element_text(size=14),
                 axis.text=element_text(size=12),
                 text=element_text(size=12),
                 legend.text=element_text(size=14),
                 strip.text=element_text(size=14)
)

idx <- c(21,41,61,81)

CP.pd.2000 <- lapply(idx, function(id) {
  sapply(raw, function(mat) mat[id,]) %>%
    {tibble(
      `calendar time`=times.cal[id],
      `event time`=times.pd,
      CP=as.numeric(rowMeans(.)))}
}) %>% bind_rows %>%
  mutate(`calendar time`=factor(`calendar time`),
         `sample size`=2000)

CP.pd.5000 <- lapply(idx, function(id) {
  sapply(raw, function(mat) mat[id,]) %>%
    {tibble(
      `calendar time`=times.cal[id],
      `event time`=times.pd,
      CP=as.numeric(rowMeans(.)))}
}) %>% bind_rows %>%
  mutate(`calendar time`=factor(`calendar time`),
         `sample size`=5000)

CP.pd.10000 <- lapply(idx, function(id) {
  sapply(raw, function(mat) mat[id,]) %>%
    {tibble(
      `calendar time`=times.cal[id],
      `event time`=times.pd,
      CP=as.numeric(rowMeans(.)))}
}) %>% bind_rows %>%
  mutate(`calendar time`=factor(`calendar time`),
         `sample size`=10000)

CP.pd <- bind_rows(CP.pd.2000, CP.pd.5000, CP.pd.10000)

CP.pd <- CP.pd %>% mutate(`event time`=`event time`*10,
                          `calendar time`=recode(`calendar time`,
                                                 `1`="10", `2`="20", `3`="30", `4`="40"))
CP.cal <- CP.cal %>% mutate(`event time`=recode(`event time`,
                                                `0.6`="6", `1.2`="12", `1.8`="18", `2.4`="24"),
                            `calendar time`=`calendar time`*10)


CPplt.pd <- CP.pd %>% mutate(`sample size`=factor(`sample size`)) %>%
  ggplot(aes(x=`event time`, linetype=`sample size`, color=`sample size`)) +
  geom_line(aes(y=CP), size=0.7) +
  geom_hline(yintercept=0.95, linetype="dotted") +
  scale_color_manual(values=c("black", "blue", "red")) +
  scale_linetype_manual(values=c("dotdash", "dotted", "solid")) +
  scale_x_continuous(expand=c(1,1)/100) +
  scale_y_continuous(name="coverage probability", expand=c(1,1)/100,
                     limits=c(0.7, 1)) +
  guides(linetype=guide_legend(nrow=1), fill=guide_legend(nrow=1)) +
  facet_grid(`calendar time`~., labeller=label_both) +
  theme_classic() + mytheme +
  theme(plot.title=element_text(hjust=0),
        legend.position="top", legend.box="horizontal")


CP.cal.2000 <- lapply(idx, function(id) {
  sapply(raw, function(mat) mat[,id]) %>%
    {tibble(
      `calendar time`=times.cal,
      `event time`=times.pd[id],
      CP=as.numeric(rowMeans(.)))}
}) %>% bind_rows %>%
  mutate(`event time`=factor(`event time`),
         `sample size`=2000)

CP.cal.5000 <- lapply(idx, function(id) {
  sapply(raw, function(mat) mat[,id]) %>%
    {tibble(
      `calendar time`=times.cal,
      `event time`=times.pd[id],
      CP=as.numeric(rowMeans(.)))}
}) %>% bind_rows %>%
  mutate(`event time`=factor(`event time`),
         `sample size`=5000)

CP.cal.10000 <- lapply(idx, function(id) {
  sapply(raw, function(mat) mat[,id]) %>%
    {tibble(
      `calendar time`=times.cal,
      `event time`=times.pd[id],
      CP=as.numeric(rowMeans(.)))}
}) %>% bind_rows %>%
  mutate(`event time`=factor(`event time`),
         `sample size`=10000)

CP.cal <- bind_rows(CP.cal.2000, CP.cal.5000, CP.cal.10000)

save(CP.pd, CP.cal, file="CP_pd_cal_pen_2000_5000_10000.RData")

CPplt.cal <- CP.cal %>% mutate(`sample size`=factor(`sample size`)) %>%
  ggplot(aes(x=`calendar time`, linetype=`sample size`, color=`sample size`)) +
  geom_line(aes(y=CP), size=0.7) +
  geom_hline(yintercept=0.95, linetype="dotted") +
  scale_color_manual(values=c("black", "blue", "red")) +
  scale_linetype_manual(values=c("dotdash", "dotted", "solid")) +
  scale_x_continuous(expand=c(1,1)/100) +
  scale_y_continuous(name=NULL, expand=c(1,1)/100,
                     limits=c(0.7, 1)) +
  guides(linetype=guide_legend(nrow=1), fill=guide_legend(nrow=1)) +
  facet_grid(`event time`~., labeller=label_both) +
  theme_classic() + mytheme +
  theme(plot.title=element_text(hjust=0),
        legend.position="top", legend.box="horizontal")

ggpubr::ggarrange(CPplt.pd, CPplt.cal, common.legend=T, ncol=2)

ggsave(file=paste0("CP_pd_cal_pen_", Sys.Date(), ".pdf"),
       width=14, height=8)
