library(tidyverse)
library(ggpubr)
cores <- 100 # number of cores in parallel
cl <- parallel::makeCluster(cores)
doParallel::registerDoParallel(cl)
`%dorng%` <- doRNG::`%dorng%`
`%dopar%` <- foreach::`%dopar%`
mytheme <- theme(plot.title=element_text(hjust=0),
                 axis.title=element_text(size=14),
                 axis.text=element_text(size=12),
                 text=element_text(size=12),
                 legend.text=element_text(size=14),
                 strip.text=element_text(size=14)
)

# sandwich estimator with chisq distribution
power <- function(mean.size.strata) {
  n.strata <- 1 # number of strata
  n.rep <- 100 # number of simulated data sets
  size.strata <- rep(mean.size.strata, n.strata) # user-specified
  rho <- 0.6 # correlation of AR(1) covar structure
  p <- 2 # number of covariates
  Z.char <- paste0("z",1:p) # covariate names
  k.pd <- 7 # number of knots in B-spline 6
  k.cal <- 7
  degree.pd <- 3 # 2
  degree.cal <- 3
  end.pd <- 3 # end-of-followup time
  end.cal <- 5
  set.seed(1234)
  level <- 0.95
  quant.upper <- qnorm((1+level)/2)
  strata <- rep(letters[1:n.strata], size.strata) # strata IDs
  times.pd <- seq(0, end.pd, end.pd/100)
  times.cal <- seq(0, end.cal, end.cal/100)
  beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
  for (i in 1:length(times.cal)) {
    for (j in 1:length(times.pd)) {
      beta.true[i, j] <- exp(-0.5*times.cal[i])*sin(3*pi*times.pd[j]/4)
    }
  }
  D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
  mu1 <- 0.005; mu2 <- 0.002
  raw <-
    foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                     .options.RNG=1234,
                     .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo", "Matrix"),
                     .combine=rbind,
                     .noexport=c("surtiver_fixtra_bresties_fit",
                                 "surtiver_fixtra_bresties_fit_diffpen")) %dopar% {
                       Rcpp::sourceCpp("surtiver_tps.cpp")
                       gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
                       Z <- mvtnorm::rmvnorm(sum(size.strata),
                                             sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
                       ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
                       U <- runif(sum(size.strata)) # survivor function F(t)
                       integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                         return(c(0.5*exp(gamma+z[1]+exp(-0.5*t.cal)*sin(3*pi*t/4)*(t<=end.pd)*z[2])))
                       }
                       time.failure <- sapply(1:sum(size.strata), function(i) {
                         res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                           integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                             log(1 - U[i]), c(0, end.pd+1)), silent=T)
                         return(ifelse(class(res)=="try-error", end.pd, res$root))
                       })
                       time.censoring <- runif(sum(size.strata), 0, end.pd)
                       data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                          time.cal=ts.cal,
                                          event=1*(time.failure<=time.censoring),
                                          strata, Z, stringsAsFactors=F)
                       names(data)[-(1:4)] <- Z.char
                       data <- data[order(data$strata, data$time.pd, -data$event),]
                       knots.pd <- quantile(data$time.pd[data$event==1],
                                            probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
                       uniqt.pd.str <- unique(data$time.pd[data$event==1])
                       bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                               knots=knots.pd, degree=degree.pd,
                                               Boundary.knots=c(0, end.pd))
                       knots.cal <- quantile(data$time.cal[data$event==1],
                                             probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
                       bases.cal <- splines::bs(data$time.cal, intercept=T,
                                                knots=knots.cal, degree=degree.cal,
                                                Boundary.knots=c(0, end.cal))
                       count.strata <- sapply(split(data$strata, data$strata), length)
                       theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
                       fit.pen <-
                         surtiver_fixtra_bresties_fit_diffpen(
                           data$event, data$time.pd, count.strata, 
                           as.matrix(data[,Z.char[2]]),
                           as.matrix(bases.cal), as.matrix(bases.pd), 
                           theta.init, as.matrix(data[,Z.char[1]]), 
                           as.numeric(0), D1, D2, mu1, mu2)
                       theta.vec.pen <- fit.pen$ctrl.pts[1,]
                       info.pen <- fit.pen$info
                       invinfo.pen <- solve(info.pen)
                       A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                                      mu1*diag(length(mu1)))
                       B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                                      mu2*diag(length(mu2)))
                       P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
                       theta.vec <- theta.vec.pen +
                         invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                         P%*%theta.vec.pen
                       info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] <-
                         info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] - P
                       mat.var <- (invinfo.pen%*%info.pen%*%invinfo.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                       Wald.test <- function(theta.vec, contrast, mat.var) {
                         pchisq(t(theta.vec)%*%t(contrast)%*%
                                  solve(contrast%*%mat.var%*%t(contrast))%*%
                                  contrast%*%theta.vec, nrow(contrast), lower=F)
                       }
                       ctrast.pd <- as.matrix(do.call(Matrix::bdiag, rep(list(diff(diag(k.pd))), k.cal)))
                       ctrast.cal <- diff(diag(k.pd*k.cal), k.pd)
                       ctrast.pdcal <- diff(diag(k.pd*k.cal))
                       test.pd <- Wald.test(theta.vec, ctrast.pd, mat.var) < 1-level
                       test.cal <- Wald.test(theta.vec, ctrast.cal, mat.var) < 1-level
                       test.pdcal <- Wald.test(theta.vec, ctrast.pdcal, mat.var) < 1-level
                       c(test.pd, test.cal, test.pdcal)
                     }
  raw <- raw[complete.cases(raw),]
  agg <- colMeans(raw)
  return(agg)
}
system.time(res.power <- sapply((1:10)*1000, power))
colnames(res.power) <- (1:10)*1000
res.power <- as_tibble(res.power) %>%
  mutate(test=c("event time", "calendar time", "event & calendar time")) %>%
  pivot_longer(-test, names_to="sample size", values_to="power") %>%
  mutate(`sample size`=as.numeric(`sample size`))
save(res.power, file="power_TPS_pen_sandwich_chisq_from1000to10000.RData")

err <- function(mean.size.strata) {
  n.strata <- 1 # number of strata
  n.rep <- 1000 # number of simulated data sets
  size.strata <- rep(mean.size.strata, n.strata) # user-specified
  rho <- 0.6 # correlation of AR(1) covar structure
  p <- 2 # number of covariates
  Z.char <- paste0("z",1:p) # covariate names
  k.pd <- 7 # number of knots in B-spline 6
  k.cal <- 7
  degree.pd <- 3 # 2
  degree.cal <- 3
  end.pd <- 3 # end-of-followup time
  end.cal <- 5
  set.seed(1234)
  level <- 0.95
  quant.upper <- qnorm((1+level)/2)
  strata <- rep(letters[1:n.strata], size.strata) # strata IDs
  times.pd <- seq(0, end.pd, end.pd/100)
  times.cal <- seq(0, end.cal, end.cal/100)
  beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
  for (i in 1:length(times.cal)) {
    for (j in 1:length(times.pd)) {
      beta.true[i, j] <- 0.5
    }
  }
  D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
  mu1 <- 0.005; mu2 <- 0.002
  raw <-
    foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                     .options.RNG=1234,
                     .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo", "Matrix"),
                     .combine=rbind,
                     .noexport=c("surtiver_fixtra_bresties_fit",
                                 "surtiver_fixtra_bresties_fit_diffpen")) %dopar% {
                       Rcpp::sourceCpp("surtiver_tps.cpp")
                       gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
                       Z <- mvtnorm::rmvnorm(sum(size.strata),
                                             sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
                       ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
                       U <- runif(sum(size.strata)) # survivor function F(t)
                       integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                         return(c(0.5*exp(gamma+z[1]+0.5*z[2])))
                       }
                       time.failure <- sapply(1:sum(size.strata), function(i) {
                         res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                           integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                             log(1 - U[i]), c(0, end.pd+1)), silent=T)
                         return(ifelse(class(res)=="try-error", end.pd, res$root))
                       })
                       time.censoring <- runif(sum(size.strata), 0, end.pd)
                       data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                          time.cal=ts.cal,
                                          event=1*(time.failure<=time.censoring),
                                          strata, Z, stringsAsFactors=F)
                       names(data)[-(1:4)] <- Z.char
                       data <- data[order(data$strata, data$time.pd, -data$event),]
                       knots.pd <- quantile(data$time.pd[data$event==1],
                                            probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
                       uniqt.pd.str <- unique(data$time.pd[data$event==1])
                       bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                               knots=knots.pd, degree=degree.pd,
                                               Boundary.knots=c(0, end.pd))
                       knots.cal <- quantile(data$time.cal[data$event==1],
                                             probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
                       bases.cal <- splines::bs(data$time.cal, intercept=T,
                                                knots=knots.cal, degree=degree.cal,
                                                Boundary.knots=c(0, end.cal))
                       count.strata <- sapply(split(data$strata, data$strata), length)
                       theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
                       fit.pen <-
                         surtiver_fixtra_bresties_fit_diffpen(
                           data$event, data$time.pd, count.strata, 
                           as.matrix(data[,Z.char[2]]),
                           as.matrix(bases.cal), as.matrix(bases.pd), 
                           theta.init, as.matrix(data[,Z.char[1]]), 
                           as.numeric(0), D1, D2, mu1, mu2)
                       theta.vec.pen <- fit.pen$ctrl.pts[1,]
                       info.pen <- fit.pen$info
                       invinfo.pen <- solve(info.pen)
                       A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                                      mu1*diag(length(mu1)))
                       B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                                      mu2*diag(length(mu2)))
                       P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
                       theta.vec <- theta.vec.pen +
                         solve(info.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                         P%*%theta.vec.pen
                       info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] <-
                         info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] - P
                       mat.var <- (invinfo.pen%*%info.pen%*%invinfo.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                       Wald.test <- function(theta.vec, contrast, mat.var) {
                         pchisq(t(theta.vec)%*%t(contrast)%*%
                                  solve(contrast%*%mat.var%*%t(contrast))%*%
                                  contrast%*%theta.vec, nrow(contrast), lower=F)
                       }
                       ctrast.pd <- as.matrix(do.call(Matrix::bdiag, rep(list(diff(diag(k.pd))), k.cal)))
                       ctrast.cal <- diff(diag(k.pd*k.cal), k.pd)
                       ctrast.pdcal <- diff(diag(k.pd*k.cal))
                       test.pd <- Wald.test(theta.vec, ctrast.pd, mat.var) < 1-level
                       test.cal <- Wald.test(theta.vec, ctrast.cal, mat.var) < 1-level
                       test.pdcal <- Wald.test(theta.vec, ctrast.pdcal, mat.var) < 1-level
                       c(test.pd, test.cal, test.pdcal)
                     }
  raw <- raw[complete.cases(raw),]
  agg <- colMeans(raw)
  return(agg)
}
system.time(res.error <- sapply((1:10)*1000, err))
colnames(res.error) <- (1:10)*1000
res.error <- as_tibble(res.error) %>%
  mutate(test=c("event time", "calendar time", "event & calendar time")) %>%
  pivot_longer(-test, names_to="sample size", values_to="type I error rate") %>%
  mutate(`sample size`=as.numeric(`sample size`))
save(res.error, file="typeIerror_TPS_pen_sandwich_chisq_from1000to10000.RData")

load("power_TPS_pen_sandwich_chisq_from1000to10000.RData")
load("typeIerror_TPS_pen_sandwich_chisq_from1000to10000.RData")

plt.err.sandwich.chisq <- res.error %>%
  mutate(test=factor(test, levels=c("event time", "calendar time", "event & calendar time"))) %>%
  ggplot(aes(x=`sample size`, group=test)) +
  geom_hline(yintercept=0.05, color="black", size=0.3, linetype="dashed") +
  geom_line(aes(y=`type I error rate`, linetype=test, color=test), size=0.6) +
  scale_x_continuous(name=NULL, expand=c(1,1)/20) +
  scale_y_continuous(name=NULL, expand=c(1,1)/100, limits=c(0,0.21)) +
  scale_color_manual(NULL, values=c("red", "blue", "green4")) +
  scale_linetype_manual(NULL, values=c("solid", "dotdash", "dotted")) +
  ggtitle("type I error rate (sandwich, chi-squared)") +
  theme_classic() +
  guides(color=guide_legend(direction="horizontal", nrow=1),
         linetype=guide_legend(direction="horizontal", nrow=1)) +
  mytheme

plt.power.sandwich.chisq <- res.power %>%
  mutate(test=factor(test, levels=c("event time", "calendar time", "event & calendar time"))) %>%
  ggplot(aes(x=`sample size`, group=test)) +
  geom_hline(yintercept=0.95, color="black", size=0.3, linetype="dashed") +
  geom_line(aes(y=power, linetype=test, color=test), size=0.6) +
  scale_x_continuous(name=NULL, expand=c(1,1)/20) +
  scale_y_continuous(name=NULL, expand=c(1,1)/20) +
  scale_color_manual(NULL, values=c("red", "blue", "green4")) +
  scale_linetype_manual(NULL, values=c("solid", "dotdash", "dotted")) +
  ggtitle("power (sandwich, chi-squared)") +
  theme_classic() +
  guides(color=guide_legend(direction="horizontal", nrow=1),
         linetype=guide_legend(direction="horizontal", nrow=1)) +
  mytheme

# model-based meat with Davies type quad form distribution
power <- function(mean.size.strata) {
  n.strata <- 1 # number of strata
  n.rep <- 100 # number of simulated data sets
  size.strata <- rep(mean.size.strata, n.strata) # user-specified
  rho <- 0.6 # correlation of AR(1) covar structure
  p <- 2 # number of covariates
  Z.char <- paste0("z",1:p) # covariate names
  k.pd <- 7 # number of knots in B-spline 6
  k.cal <- 7
  degree.pd <- 3 # 2
  degree.cal <- 3
  end.pd <- 3 # end-of-followup time
  end.cal <- 5
  set.seed(1234)
  level <- 0.95
  quant.upper <- qnorm((1+level)/2)
  strata <- rep(letters[1:n.strata], size.strata) # strata IDs
  times.pd <- seq(0, end.pd, end.pd/100)
  times.cal <- seq(0, end.cal, end.cal/100)
  beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
  for (i in 1:length(times.cal)) {
    for (j in 1:length(times.pd)) {
      beta.true[i, j] <- exp(-0.5*times.cal[i])*sin(3*pi*times.pd[j]/4)
    }
  }
  D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
  mu1 <- 0.005; mu2 <- 0.002
  raw <-
    foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                     .options.RNG=1234,
                     .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo", "Matrix"),
                     .combine=rbind,
                     .noexport=c("surtiver_fixtra_bresties_fit",
                                 "surtiver_fixtra_bresties_fit_diffpen")) %dopar% {
                                   Rcpp::sourceCpp("surtiver_tps.cpp")
                                   gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
                                   Z <- mvtnorm::rmvnorm(sum(size.strata),
                                                         sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
                                   ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
                                   U <- runif(sum(size.strata)) # survivor function F(t)
                                   integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                                     return(c(0.5*exp(gamma+z[1]+exp(-0.5*t.cal)*sin(3*pi*t/4)*(t<=end.pd)*z[2])))
                                   }
                                   time.failure <- sapply(1:sum(size.strata), function(i) {
                                     res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                                       integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                                         log(1 - U[i]), c(0, end.pd+1)), silent=T)
                                     return(ifelse(class(res)=="try-error", end.pd, res$root))
                                   })
                                   time.censoring <- runif(sum(size.strata), 0, end.pd)
                                   data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                                      time.cal=ts.cal,
                                                      event=1*(time.failure<=time.censoring),
                                                      strata, Z, stringsAsFactors=F)
                                   names(data)[-(1:4)] <- Z.char
                                   data <- data[order(data$strata, data$time.pd, -data$event),]
                                   knots.pd <- quantile(data$time.pd[data$event==1],
                                                        probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
                                   uniqt.pd.str <- unique(data$time.pd[data$event==1])
                                   bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                                           knots=knots.pd, degree=degree.pd,
                                                           Boundary.knots=c(0, end.pd))
                                   knots.cal <- quantile(data$time.cal[data$event==1],
                                                         probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
                                   bases.cal <- splines::bs(data$time.cal, intercept=T,
                                                            knots=knots.cal, degree=degree.cal,
                                                            Boundary.knots=c(0, end.cal))
                                   count.strata <- sapply(split(data$strata, data$strata), length)
                                   theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
                                   fit.pen <-
                                     surtiver_fixtra_bresties_fit_diffpen(
                                       data$event, data$time.pd, count.strata, 
                                       as.matrix(data[,Z.char[2]]),
                                       as.matrix(bases.cal), as.matrix(bases.pd), 
                                       theta.init, as.matrix(data[,Z.char[1]]), 
                                       as.numeric(0), D1, D2, mu1, mu2)
                                   theta.vec.pen <- fit.pen$ctrl.pts[1,]
                                   info.pen <- fit.pen$info
                                   invinfo.pen <- solve(info.pen)
                                   A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                                                  mu1*diag(length(mu1)))
                                   B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                                                  mu2*diag(length(mu2)))
                                   P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
                                   theta.vec <- theta.vec.pen +
                                     solve(info.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                                     P%*%theta.vec.pen
                                   mat.quad <- invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] <-
                                     info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] - P
                                   mat.var <- (invinfo.pen%*%info.pen%*%invinfo.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   Wald.test <- function(theta.vec, contrast, mat.var, mat.quad) {
                                     meat <- solve(contrast%*%mat.quad%*%t(contrast))
                                     lambdas <- 
                                       eigen(meat%*%contrast%*%mat.var%*%t(contrast))$values
                                     q <- t(theta.vec)%*%t(contrast)%*%meat%*%contrast%*%theta.vec
                                     CompQuadForm::davies(q, lambdas)$Qq
                                   }
                                   ctrast.pd <- as.matrix(do.call(Matrix::bdiag, rep(list(diff(diag(k.pd))), k.cal)))
                                   ctrast.cal <- diff(diag(k.pd*k.cal), k.pd)
                                   ctrast.pdcal <- diff(diag(k.pd*k.cal))
                                   test.pd <- Wald.test(theta.vec, ctrast.pd, mat.var, mat.quad) < 1-level
                                   test.cal <- Wald.test(theta.vec, ctrast.cal, mat.var, mat.quad) < 1-level
                                   test.pdcal <- Wald.test(theta.vec, ctrast.pdcal, mat.var, mat.quad) < 1-level
                                   c(test.pd, test.cal, test.pdcal)
                                 }
  raw <- raw[complete.cases(raw),]
  agg <- colMeans(raw)
  return(agg)
}
system.time(res.power <- sapply((1:10)*1000, power))
colnames(res.power) <- (1:10)*1000
res.power <- as_tibble(res.power) %>%
  mutate(test=c("event time", "calendar time", "event & calendar time")) %>%
  pivot_longer(-test, names_to="sample size", values_to="power") %>%
  mutate(`sample size`=as.numeric(`sample size`))
save(res.power, file="power_TPS_pen_model_davies_from1000to10000.RData")


err <- function(mean.size.strata) {
  n.strata <- 1 # number of strata
  n.rep <- 1000 # number of simulated data sets
  size.strata <- rep(mean.size.strata, n.strata) # user-specified
  rho <- 0.6 # correlation of AR(1) covar structure
  p <- 2 # number of covariates
  Z.char <- paste0("z",1:p) # covariate names
  k.pd <- 7 # number of knots in B-spline 6
  k.cal <- 7
  degree.pd <- 3 # 2
  degree.cal <- 3
  end.pd <- 3 # end-of-followup time
  end.cal <- 5
  set.seed(1234)
  level <- 0.95
  quant.upper <- qnorm((1+level)/2)
  strata <- rep(letters[1:n.strata], size.strata) # strata IDs
  times.pd <- seq(0, end.pd, end.pd/100)
  times.cal <- seq(0, end.cal, end.cal/100)
  beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
  for (i in 1:length(times.cal)) {
    for (j in 1:length(times.pd)) {
      beta.true[i, j] <- 0.5
    }
  }
  D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
  mu1 <- 0.005; mu2 <- 0.002
  raw <-
    foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                     .options.RNG=1234,
                     .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo", "Matrix", "CompQuadForm"),
                     .combine=rbind,
                     .noexport=c("surtiver_fixtra_bresties_fit",
                                 "surtiver_fixtra_bresties_fit_diffpen")) %dopar% {
                                   Rcpp::sourceCpp("surtiver_tps.cpp")
                                   gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
                                   Z <- mvtnorm::rmvnorm(sum(size.strata),
                                                         sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
                                   ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
                                   U <- runif(sum(size.strata)) # survivor function F(t)
                                   integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                                     return(c(0.5*exp(gamma+z[1]+0.5*z[2])))
                                   }
                                   time.failure <- sapply(1:sum(size.strata), function(i) {
                                     res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                                       integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                                         log(1 - U[i]), c(0, end.pd+1)), silent=T)
                                     return(ifelse(class(res)=="try-error", end.pd, res$root))
                                   })
                                   time.censoring <- runif(sum(size.strata), 0, end.pd)
                                   data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                                      time.cal=ts.cal,
                                                      event=1*(time.failure<=time.censoring),
                                                      strata, Z, stringsAsFactors=F)
                                   names(data)[-(1:4)] <- Z.char
                                   data <- data[order(data$strata, data$time.pd, -data$event),]
                                   knots.pd <- quantile(data$time.pd[data$event==1],
                                                        probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
                                   uniqt.pd.str <- unique(data$time.pd[data$event==1])
                                   bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                                           knots=knots.pd, degree=degree.pd,
                                                           Boundary.knots=c(0, end.pd))
                                   knots.cal <- quantile(data$time.cal[data$event==1],
                                                         probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
                                   bases.cal <- splines::bs(data$time.cal, intercept=T,
                                                            knots=knots.cal, degree=degree.cal,
                                                            Boundary.knots=c(0, end.cal))
                                   count.strata <- sapply(split(data$strata, data$strata), length)
                                   theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
                                   fit.pen <-
                                     surtiver_fixtra_bresties_fit_diffpen(
                                       data$event, data$time.pd, count.strata, 
                                       as.matrix(data[,Z.char[2]]),
                                       as.matrix(bases.cal), as.matrix(bases.pd), 
                                       theta.init, as.matrix(data[,Z.char[1]]), 
                                       as.numeric(0), D1, D2, mu1, mu2)
                                   theta.vec.pen <- fit.pen$ctrl.pts[1,]
                                   info.pen <- fit.pen$info
                                   invinfo.pen <- solve(info.pen)
                                   A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                                                  mu1*diag(length(mu1)))
                                   B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                                                  mu2*diag(length(mu2)))
                                   P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
                                   theta.vec <- theta.vec.pen +
                                     solve(info.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                                     P%*%theta.vec.pen
                                   mat.quad <- invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] <-
                                     info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] - P
                                   mat.var <- (invinfo.pen%*%info.pen%*%invinfo.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   Wald.test <- function(theta.vec, contrast, mat.var, mat.quad) {
                                     meat <- solve(contrast%*%mat.quad%*%t(contrast))
                                     lambdas <- 
                                       eigen(meat%*%contrast%*%mat.var%*%t(contrast))$values
                                     q <- t(theta.vec)%*%t(contrast)%*%meat%*%contrast%*%theta.vec
                                     CompQuadForm::davies(q, lambdas)$Qq
                                   }
                                   ctrast.pd <- as.matrix(do.call(Matrix::bdiag, rep(list(diff(diag(k.pd))), k.cal)))
                                   ctrast.cal <- diff(diag(k.pd*k.cal), k.pd)
                                   ctrast.pdcal <- diff(diag(k.pd*k.cal))
                                   test.pd <- Wald.test(theta.vec, ctrast.pd, mat.var, mat.quad) < 1-level
                                   test.cal <- Wald.test(theta.vec, ctrast.cal, mat.var, mat.quad) < 1-level
                                   test.pdcal <- Wald.test(theta.vec, ctrast.pdcal, mat.var, mat.quad) < 1-level
                                   c(test.pd, test.cal, test.pdcal)
                                 }
  raw <- raw[complete.cases(raw),]
  agg <- colMeans(raw)
  return(agg)
}
system.time(res.error <- sapply((1:10)*1000, err))
colnames(res.error) <- (1:10)*1000
res.error <- as_tibble(res.error) %>%
  mutate(test=c("event time", "calendar time", "event & calendar time")) %>%
  pivot_longer(-test, names_to="sample size", values_to="type I error rate") %>%
  mutate(`sample size`=as.numeric(`sample size`))
save(res.error, file="typeIerror_TPS_pen_model_davies_from1000to10000.RData")

load("power_TPS_pen_model_davies_from1000to10000.RData")
load("typeIerror_TPS_pen_model_davies_from1000to10000.RData")

plt.err.model.davies <- res.error %>%
  mutate(test=factor(test, levels=c("event time", "calendar time", "event & calendar time"))) %>%
  ggplot(aes(x=`sample size`, group=test)) +
  geom_hline(yintercept=0.05, color="black", size=0.3, linetype="dashed") +
  geom_line(aes(y=`type I error rate`, linetype=test, color=test), size=0.6) +
  scale_x_continuous(name=NULL, expand=c(1,1)/20) +
  scale_y_continuous(name=NULL, expand=c(1,1)/100, limits=c(0,0.21)) +
  scale_color_manual(NULL, values=c("red", "blue", "green4")) +
  scale_linetype_manual(NULL, values=c("solid", "dotdash", "dotted")) +
  ggtitle("type I error rate (model-based, Davies)") +
  theme_classic() +
  guides(color=guide_legend(direction="horizontal", nrow=1),
         linetype=guide_legend(direction="horizontal", nrow=1)) +
  mytheme

plt.power.model.davies <- res.power %>%
  mutate(test=factor(test, levels=c("event time", "calendar time", "event & calendar time"))) %>%
  ggplot(aes(x=`sample size`, group=test)) +
  geom_hline(yintercept=0.95, color="black", size=0.3, linetype="dashed") +
  geom_line(aes(y=power, linetype=test, color=test), size=0.6) +
  scale_x_continuous(name=NULL, expand=c(1,1)/20) +
  scale_y_continuous(name=NULL, expand=c(1,1)/20) +
  scale_color_manual(NULL, values=c("red", "blue", "green4")) +
  scale_linetype_manual(NULL, values=c("solid", "dotdash", "dotted")) +
  ggtitle("power (model-based, Davies)") +
  theme_classic() +
  guides(color=guide_legend(direction="horizontal", nrow=1),
         linetype=guide_legend(direction="horizontal", nrow=1)) +
  mytheme


# model-based meat with chisq distribution
power <- function(mean.size.strata) {
  n.strata <- 1 # number of strata
  n.rep <- 100 # number of simulated data sets
  size.strata <- rep(mean.size.strata, n.strata) # user-specified
  rho <- 0.6 # correlation of AR(1) covar structure
  p <- 2 # number of covariates
  Z.char <- paste0("z",1:p) # covariate names
  k.pd <- 7 # number of knots in B-spline 6
  k.cal <- 7
  degree.pd <- 3 # 2
  degree.cal <- 3
  end.pd <- 3 # end-of-followup time
  end.cal <- 5
  set.seed(1234)
  level <- 0.95
  quant.upper <- qnorm((1+level)/2)
  strata <- rep(letters[1:n.strata], size.strata) # strata IDs
  times.pd <- seq(0, end.pd, end.pd/100)
  times.cal <- seq(0, end.cal, end.cal/100)
  beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
  for (i in 1:length(times.cal)) {
    for (j in 1:length(times.pd)) {
      beta.true[i, j] <- exp(-0.5*times.cal[i])*sin(3*pi*times.pd[j]/4)
    }
  }
  D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
  mu1 <- 0.005; mu2 <- 0.002
  raw <-
    foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                     .options.RNG=1234,
                     .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo", "Matrix"),
                     .combine=rbind,
                     .noexport=c("surtiver_fixtra_bresties_fit",
                                 "surtiver_fixtra_bresties_fit_diffpen")) %dopar% {
                                   Rcpp::sourceCpp("surtiver_tps.cpp")
                                   gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
                                   Z <- mvtnorm::rmvnorm(sum(size.strata),
                                                         sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
                                   ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
                                   U <- runif(sum(size.strata)) # survivor function F(t)
                                   integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                                     return(c(0.5*exp(gamma+z[1]+exp(-0.5*t.cal)*sin(3*pi*t/4)*(t<=end.pd)*z[2])))
                                   }
                                   time.failure <- sapply(1:sum(size.strata), function(i) {
                                     res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                                       integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                                         log(1 - U[i]), c(0, end.pd+1)), silent=T)
                                     return(ifelse(class(res)=="try-error", end.pd, res$root))
                                   })
                                   time.censoring <- runif(sum(size.strata), 0, end.pd)
                                   data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                                      time.cal=ts.cal,
                                                      event=1*(time.failure<=time.censoring),
                                                      strata, Z, stringsAsFactors=F)
                                   names(data)[-(1:4)] <- Z.char
                                   data <- data[order(data$strata, data$time.pd, -data$event),]
                                   knots.pd <- quantile(data$time.pd[data$event==1],
                                                        probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
                                   uniqt.pd.str <- unique(data$time.pd[data$event==1])
                                   bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                                           knots=knots.pd, degree=degree.pd,
                                                           Boundary.knots=c(0, end.pd))
                                   knots.cal <- quantile(data$time.cal[data$event==1],
                                                         probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
                                   bases.cal <- splines::bs(data$time.cal, intercept=T,
                                                            knots=knots.cal, degree=degree.cal,
                                                            Boundary.knots=c(0, end.cal))
                                   count.strata <- sapply(split(data$strata, data$strata), length)
                                   theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
                                   fit.pen <-
                                     surtiver_fixtra_bresties_fit_diffpen(
                                       data$event, data$time.pd, count.strata, 
                                       as.matrix(data[,Z.char[2]]),
                                       as.matrix(bases.cal), as.matrix(bases.pd), 
                                       theta.init, as.matrix(data[,Z.char[1]]), 
                                       as.numeric(0), D1, D2, mu1, mu2)
                                   theta.vec.pen <- fit.pen$ctrl.pts[1,]
                                   info.pen <- fit.pen$info
                                   invinfo.pen <- solve(info.pen)
                                   A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                                                  mu1*diag(length(mu1)))
                                   B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                                                  mu2*diag(length(mu2)))
                                   P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
                                   theta.vec <- theta.vec.pen +
                                     solve(info.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                                     P%*%theta.vec.pen
                                   mat.var <- invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   Wald.test <- function(theta.vec, contrast, mat.var) {
                                     pchisq(t(theta.vec)%*%t(contrast)%*%
                                              solve(contrast%*%mat.var%*%t(contrast))%*%
                                              contrast%*%theta.vec, nrow(contrast), lower=F)
                                   }
                                   ctrast.pd <- as.matrix(do.call(Matrix::bdiag, rep(list(diff(diag(k.pd))), k.cal)))
                                   ctrast.cal <- diff(diag(k.pd*k.cal), k.pd)
                                   ctrast.pdcal <- diff(diag(k.pd*k.cal))
                                   test.pd <- Wald.test(theta.vec, ctrast.pd, mat.var) < 1-level
                                   test.cal <- Wald.test(theta.vec, ctrast.cal, mat.var) < 1-level
                                   test.pdcal <- Wald.test(theta.vec, ctrast.pdcal, mat.var) < 1-level
                                   c(test.pd, test.cal, test.pdcal)
                                 }
  raw <- raw[complete.cases(raw),]
  agg <- colMeans(raw)
  return(agg)
}
system.time(res.power <- sapply((1:10)*1000, power))
colnames(res.power) <- (1:10)*1000
res.power <- as_tibble(res.power) %>%
  mutate(test=c("event time", "calendar time", "event & calendar time")) %>%
  pivot_longer(-test, names_to="sample size", values_to="power") %>%
  mutate(`sample size`=as.numeric(`sample size`))
save(res.power, file="power_TPS_pen_model_chisq_from1000to10000.RData")


err <- function(mean.size.strata) {
  n.strata <- 1 # number of strata
  n.rep <- 1000 # number of simulated data sets
  size.strata <- rep(mean.size.strata, n.strata) # user-specified
  rho <- 0.6 # correlation of AR(1) covar structure
  p <- 2 # number of covariates
  Z.char <- paste0("z",1:p) # covariate names
  k.pd <- 7 # number of knots in B-spline 6
  k.cal <- 7
  degree.pd <- 3 # 2
  degree.cal <- 3
  end.pd <- 3 # end-of-followup time
  end.cal <- 5
  set.seed(1234)
  level <- 0.95
  quant.upper <- qnorm((1+level)/2)
  strata <- rep(letters[1:n.strata], size.strata) # strata IDs
  times.pd <- seq(0, end.pd, end.pd/100)
  times.cal <- seq(0, end.cal, end.cal/100)
  beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
  for (i in 1:length(times.cal)) {
    for (j in 1:length(times.pd)) {
      beta.true[i, j] <- 0.5
    }
  }
  D1 <- diff(diag(k.cal)); D2 <- diff(diag(k.pd))
  mu1 <- 0.005; mu2 <- 0.002
  raw <-
    foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                     .options.RNG=1234,
                     .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo", "Matrix", "CompQuadForm"),
                     .combine=rbind,
                     .noexport=c("surtiver_fixtra_bresties_fit",
                                 "surtiver_fixtra_bresties_fit_diffpen")) %dopar% {
                                   Rcpp::sourceCpp("surtiver_tps.cpp")
                                   gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
                                   Z <- mvtnorm::rmvnorm(sum(size.strata),
                                                         sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
                                   ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
                                   U <- runif(sum(size.strata)) # survivor function F(t)
                                   integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                                     return(c(0.5*exp(gamma+z[1]+0.5*z[2])))
                                   }
                                   time.failure <- sapply(1:sum(size.strata), function(i) {
                                     res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                                       integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                                         log(1 - U[i]), c(0, end.pd+1)), silent=T)
                                     return(ifelse(class(res)=="try-error", end.pd, res$root))
                                   })
                                   time.censoring <- runif(sum(size.strata), 0, end.pd)
                                   data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                                      time.cal=ts.cal,
                                                      event=1*(time.failure<=time.censoring),
                                                      strata, Z, stringsAsFactors=F)
                                   names(data)[-(1:4)] <- Z.char
                                   data <- data[order(data$strata, data$time.pd, -data$event),]
                                   knots.pd <- quantile(data$time.pd[data$event==1],
                                                        probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
                                   uniqt.pd.str <- unique(data$time.pd[data$event==1])
                                   bases.pd <- splines::bs(uniqt.pd.str, intercept=T,
                                                           knots=knots.pd, degree=degree.pd,
                                                           Boundary.knots=c(0, end.pd))
                                   knots.cal <- quantile(data$time.cal[data$event==1],
                                                         probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
                                   bases.cal <- splines::bs(data$time.cal, intercept=T,
                                                            knots=knots.cal, degree=degree.cal,
                                                            Boundary.knots=c(0, end.cal))
                                   count.strata <- sapply(split(data$strata, data$strata), length)
                                   theta.init <- matrix(runif(1*k.pd*k.cal), 1, k.pd*k.cal)
                                   fit.pen <-
                                     surtiver_fixtra_bresties_fit_diffpen(
                                       data$event, data$time.pd, count.strata, 
                                       as.matrix(data[,Z.char[2]]),
                                       as.matrix(bases.cal), as.matrix(bases.pd), 
                                       theta.init, as.matrix(data[,Z.char[1]]), 
                                       as.numeric(0), D1, D2, mu1, mu2)
                                   theta.vec.pen <- fit.pen$ctrl.pts[1,]
                                   info.pen <- fit.pen$info
                                   invinfo.pen <- solve(info.pen)
                                   A <- kronecker(kronecker(D1, diag(ncol(bases.pd))),
                                                  mu1*diag(length(mu1)))
                                   B <- kronecker(kronecker(diag(ncol(bases.cal)), D2),
                                                  mu2*diag(length(mu2)))
                                   P <- (nrow(data)^0.25)*(t(A)%*%A + t(B)%*%B)
                                   theta.vec <- theta.vec.pen +
                                     solve(info.pen)[1:(k.cal*k.pd), 1:(k.cal*k.pd)]%*%
                                     P%*%theta.vec.pen
                                   mat.quad <- invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] <-
                                     info.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)] - P
                                   mat.var <- invinfo.pen[1:(k.cal*k.pd), 1:(k.cal*k.pd)]
                                   Wald.test <- function(theta.vec, contrast, mat.var) {
                                     pchisq(t(theta.vec)%*%t(contrast)%*%
                                              solve(contrast%*%mat.var%*%t(contrast))%*%
                                              contrast%*%theta.vec, nrow(contrast), lower=F)
                                   }
                                   ctrast.pd <- as.matrix(do.call(Matrix::bdiag, rep(list(diff(diag(k.pd))), k.cal)))
                                   ctrast.cal <- diff(diag(k.pd*k.cal), k.pd)
                                   ctrast.pdcal <- diff(diag(k.pd*k.cal))
                                   test.pd <- Wald.test(theta.vec, ctrast.pd, mat.var) < 1-level
                                   test.cal <- Wald.test(theta.vec, ctrast.cal, mat.var) < 1-level
                                   test.pdcal <- Wald.test(theta.vec, ctrast.pdcal, mat.var) < 1-level
                                   c(test.pd, test.cal, test.pdcal)
                                 }
  raw <- raw[complete.cases(raw),]
  agg <- colMeans(raw)
  return(agg)
}
system.time(res.error <- sapply((1:10)*1000, err))
colnames(res.error) <- (1:10)*1000
res.error <- as_tibble(res.error) %>%
  mutate(test=c("event time", "calendar time", "event & calendar time")) %>%
  pivot_longer(-test, names_to="sample size", values_to="type I error rate") %>%
  mutate(`sample size`=as.numeric(`sample size`))
save(res.error, file="typeIerror_TPS_pen_model_chisq_from1000to10000.RData")

load("power_TPS_pen_model_chisq_from1000to10000.RData")
load("typeIerror_TPS_pen_model_chisq_from1000to10000.RData")

plt.err.model.chisq <- res.error %>%
  mutate(test=factor(test, levels=c("event time", "calendar time", "event & calendar time"))) %>%
  ggplot(aes(x=`sample size`, group=test)) +
  geom_hline(yintercept=0.05, color="black", size=0.3, linetype="dashed") +
  geom_line(aes(y=`type I error rate`, linetype=test, color=test), size=0.6) +
  scale_x_continuous(name=NULL, expand=c(1,1)/20) +
  scale_y_continuous(name=NULL, expand=c(1,1)/100, limits=c(0,0.21)) +
  scale_color_manual(NULL, values=c("red", "blue", "green4")) +
  scale_linetype_manual(NULL, values=c("solid", "dotdash", "dotted")) +
  ggtitle("type I error rate (model-based, chi-squared)") +
  theme_classic() +
  guides(color=guide_legend(direction="horizontal", nrow=1),
         linetype=guide_legend(direction="horizontal", nrow=1)) +
  mytheme

plt.power.model.chisq <- res.power %>%
  mutate(test=factor(test, levels=c("event time", "calendar time", "event & calendar time"))) %>%
  ggplot(aes(x=`sample size`, group=test)) +
  geom_hline(yintercept=0.95, color="black", size=0.3, linetype="dashed") +
  geom_line(aes(y=power, linetype=test, color=test), size=0.6) +
  scale_x_continuous(name=NULL, expand=c(1,1)/20) +
  scale_y_continuous(name=NULL, expand=c(1,1)/20) +
  scale_color_manual(NULL, values=c("red", "blue", "green4")) +
  scale_linetype_manual(NULL, values=c("solid", "dotdash", "dotted")) +
  ggtitle("power (model-based, chi-squared)") +
  theme_classic() +
  guides(color=guide_legend(direction="horizontal", nrow=1),
         linetype=guide_legend(direction="horizontal", nrow=1)) +
  mytheme

ggpubr::annotate_figure(
  ggpubr::ggarrange(plt.err.sandwich.chisq, plt.err.model.chisq, plt.err.model.davies,
                    plt.power.sandwich.chisq, plt.power.model.chisq, plt.power.model.davies,
                    nrow=2, ncol=3, common.legend=T),
  bottom=text_grob("sample size", size=14))
ggsave(file=paste0("errpower_pd_cal_pen_", Sys.Date(), ".pdf"),
       width=14, height=8)
parallel::stopCluster(cl)