library(tidyverse)
library(ggpubr)
n.rep <- 100 # number of simulated data sets
cores <- 100 # number of cores in parallel
n.strata <- 1 # number of strata
mean.size.strata <- 3000 # average size of strata
# size.strata <- rpois(n.strata, mean.size.strata) # can be user-specified
size.strata <- rep(mean.size.strata, n.strata) # user-specified
rho <- 0.6 # correlation of AR(1) covar structure
p <- 2 # number of covariates
Z.char <- paste0("z",1:p) # covariate names
k.pd <- 7 # number of knots in B-spline 6
k.cal <- 7
degree.pd <- 3 # 2
degree.cal <- 3
end.pd <- 3 # end-of-followup time
end.cal <- 5
set.seed(1234)
level <- 0.95
quant.upper <- qnorm((1+level)/2)
strata <- rep(letters[1:n.strata], size.strata) # strata IDs
times.pd <- seq(0, end.pd, end.pd/100)
times.cal <- seq(0, end.cal, end.cal/100)
beta.true <- matrix(nrow=length(times.cal), ncol=length(times.pd))
for (i in 1:length(times.cal)) {
  for (j in 1:length(times.pd)) {
    beta.true[i, j] <- exp(-0.5*times.cal[i])*sin(3*pi*times.pd[j]/4)
  }
}
cl <- parallel::makeCluster(cores)
doParallel::registerDoParallel(cl)
`%dorng%` <- doRNG::`%dorng%`
`%dopar%` <- foreach::`%dopar%`
system.time(raw <-
              foreach::foreach(i=1:n.rep, .verbose=F, #.errorhandling="remove", 
                               .options.RNG=1234,
                               .packages=c("mvtnorm","survival","Rcpp","RcppArmadillo",
                                           "tidyverse"),
                               .combine=rbind,
                               .noexport=c("surtiver_fixtra_bresties_fit_diffpen",
                                           "surtiver_fixtra_bresties_fit",
                                           "sumDRsq_fixtra_bresties_cv_DR",
                                           "logplkd_fixtra_bresties_cv_UC",
                                           "logplkd_fixtra_bresties_standalone"))
            %dorng% {
              Rcpp::sourceCpp("surtiver_tps.cpp")
              source("surtiver_tps_cv.R")
              gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
              Z <- mvtnorm::rmvnorm(sum(size.strata),
                                    sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
              ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
              U <- runif(sum(size.strata)) # survivor function F(t)
              integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                return(c(0.5*exp(gamma+z[1]+exp(-0.5*t.cal)*sin(3*pi*t/4)*(t<=end.pd)*z[2])))
              }
              time.failure <- sapply(1:sum(size.strata), function(i) {
                res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                  integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                    log(1 - U[i]), c(0, end.pd+1)), silent=T)
                return(ifelse(class(res)=="try-error", end.pd, res$root))
              })
              time.censoring <- runif(sum(size.strata), 0, end.pd)
              data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                 time.cal=ts.cal,
                                 event=1*(time.failure<=time.censoring),
                                 strata, Z, stringsAsFactors=F)
              names(data)[-(1:4)] <- Z.char
              knots.pd <- quantile(data$time.pd[data$event==1],
                                   probs=(1:(k.pd-degree.pd-1))/(k.pd-degree.pd))
              knots.cal <- quantile(data$time.cal[data$event==1],
                                    probs=(1:(k.cal-degree.cal-1))/(k.cal-degree.cal))
              B.cal <- as.matrix(splines::bs(times.cal, degree=degree.cal,
                                             intercept=T, knots=knots.cal,
                                             Boundary.knots=c(0, end.cal)))
              B.pd <- splines::bs(times.pd, degree=degree.pd, intercept=T,
                                  knots=knots.pd, Boundary.knots=c(0, end.pd))
              # UC
              system.time(cv.UC <-
                            cv.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                        "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                                        degree.time=degree.pd, degree.var=degree.cal,
                                        nfold=4, method="UC", parallel=F))
              fit.cv.UC <-
                surbiver(data, "event", "time.pd", "time.cal", "strata",
                         "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                         degree.time=degree.pd, degree.var=degree.cal, pen=T,
                         lambda.time=cv.UC %>% slice_min(cve) %>% pull(lambda.time),
                         lambda.var=cv.UC %>% slice_min(cve) %>% pull(lambda.var))
              negunpenllkd.UC <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.UC$ctrl.pts, fit.cv.UC$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              beta.UC <- B.cal%*%matrix(fit.cv.UC$ctrl.pts[1,], byrow=T,
                                         ncol=k.pd)%*%t(B.pd)
              IMSE.UC <- mean((beta.UC-beta.true)^2)
              # CFC
              system.time(cv.CFC <-
                            cv.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                        "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                                        degree.time=degree.pd, degree.var=degree.cal,
                                        nfold=4, method="CFC", parallel=F))
              fit.cv.CFC <-
                surbiver(data, "event", "time.pd", "time.cal", "strata",
                         "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                         degree.time=degree.pd, degree.var=degree.cal, pen=T,
                         lambda.time=cv.CFC %>% slice_min(cve) %>% pull(lambda.time),
                         lambda.var=cv.CFC %>% slice_min(cve) %>% pull(lambda.var))
              negunpenllkd.CFC <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.CFC$ctrl.pts, fit.cv.CFC$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              beta.CFC <- B.cal%*%matrix(fit.cv.CFC$ctrl.pts[1,], byrow=T,
                                        ncol=k.pd)%*%t(B.pd)
              IMSE.CFC <- mean((beta.CFC-beta.true)^2)
              # FC
              system.time(cv.FC <-
                            cv.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                        "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                                        degree.time=degree.pd, degree.var=degree.cal,
                                        nfold=4, method="FC", parallel=F))
              fit.cv.FC <-
                surbiver(data, "event", "time.pd", "time.cal", "strata",
                         "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                         degree.time=degree.pd, degree.var=degree.cal, pen=T,
                         lambda.time=cv.FC %>% slice_min(cve) %>% pull(lambda.time),
                         lambda.var=cv.FC %>% slice_min(cve) %>% pull(lambda.var))
              negunpenllkd.FC <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.FC$ctrl.pts, fit.cv.FC$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              beta.FC <- B.cal%*%matrix(fit.cv.FC$ctrl.pts[1,], byrow=T,
                                         ncol=k.pd)%*%t(B.pd)
              IMSE.FC <- mean((beta.FC-beta.true)^2)
              # DR
              system.time(cv.DR <-
                            cv.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                        "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                                        degree.time=degree.pd, degree.var=degree.cal, 
                                        nfold=4, method="DR", parallel=F))
              fit.cv.DR <-
                surbiver(data, "event", "time.pd", "time.cal", "strata",
                         "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                         degree.time=degree.pd, degree.var=degree.cal, pen=T,
                         lambda.time=cv.DR %>% slice_min(cve) %>% pull(lambda.time),
                         lambda.var=cv.DR %>% slice_min(cve) %>% pull(lambda.var))
              negunpenllkd.DR <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.DR$ctrl.pts, fit.cv.DR$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              beta.DR <- B.cal%*%matrix(fit.cv.DR$ctrl.pts[1,], byrow=T,
                                        ncol=k.pd)%*%t(B.pd)
              IMSE.DR <- mean((beta.DR-beta.true)^2)
              # GCV
              system.time(cv.GCV <-
                            cv.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                        "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                                        degree.time=degree.pd, degree.var=degree.cal,
                                        nfold=4, method="GCV"))
              fit.cv.GCV <-
                surbiver(data, "event", "time.pd", "time.cal", "strata",
                         "z2", "z1", nbs.time=k.pd, nbs.var=k.cal,
                         degree.time=degree.pd, degree.var=degree.cal, pen=T,
                         lambda.time=cv.GCV %>% slice_min(cve) %>% pull(lambda.time),
                         lambda.var=cv.GCV %>% slice_min(cve) %>% pull(lambda.var))
              negunpenllkd.GCV <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.GCV$ctrl.pts, fit.cv.GCV$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              beta.GCV <- B.cal%*%matrix(fit.cv.GCV$ctrl.pts[1,], byrow=T,
                                        ncol=k.pd)%*%t(B.pd)
              IMSE.GCV <- mean((beta.GCV-beta.true)^2)
              
              # validation data
              gamma <- rep(rnorm(n.strata, sd=0.5), size.strata) # strata-specific effects
              Z <- mvtnorm::rmvnorm(sum(size.strata),
                                    sigma=rho^abs(outer(1:p, 1:p, "-"))) # covariates
              ts.cal <- round(runif(sum(size.strata), 0, end.cal), 5)
              U <- runif(sum(size.strata)) # survivor function F(t)
              integrand <- function(t, gamma, z, t.cal) { # gamma is a scalar, z is a row of Z
                return(c(0.5*exp(gamma+z[1]+exp(-0.5*t.cal)*sin(3*pi*t/4)*(t<=end.pd)*z[2])))
              }
              time.failure <- sapply(1:sum(size.strata), function(i) {
                res <- try(uniroot(function(t) integrate(Vectorize(function(x)
                  integrand(x, gamma[i], Z[i,], ts.cal[i])), 0, t)$value +
                    log(1 - U[i]), c(0, end.pd+1)), silent=T)
                return(ifelse(class(res)=="try-error", end.pd, res$root))
              })
              time.censoring <- runif(sum(size.strata), 0, end.pd)
              data <- data.frame(time.pd=round(pmin(time.failure, time.censoring), 5),
                                 time.cal=ts.cal,
                                 event=1*(time.failure<=time.censoring),
                                 strata, Z, stringsAsFactors=F)
              names(data)[-(1:4)] <- Z.char
              # performance on validation data
              negunpenllkd.UC.val <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.UC$ctrl.pts, fit.cv.UC$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              negunpenllkd.CFC.val <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.CFC$ctrl.pts, fit.cv.CFC$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              negunpenllkd.FC.val <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.FC$ctrl.pts, fit.cv.FC$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              negunpenllkd.DR.val <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.DR$ctrl.pts, fit.cv.DR$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              negunpenllkd.GCV.val <-
                -2*unpenllkd.surbiver(data, "event", "time.pd", "time.cal", "strata",
                                      "z2", "z1", fit.cv.GCV$ctrl.pts, fit.cv.GCV$tief,
                                      nbs.time=k.pd, nbs.var=k.cal,
                                      degree.time=degree.pd, degree.var=degree.cal)
              c(negunpenllkd.FC, negunpenllkd.CFC, negunpenllkd.UC, negunpenllkd.DR, negunpenllkd.GCV,
                negunpenllkd.FC.val, negunpenllkd.CFC.val, negunpenllkd.UC.val,
                negunpenllkd.DR.val, negunpenllkd.GCV.val,
                IMSE.FC, IMSE.CFC, IMSE.UC, IMSE.DR, IMSE.GCV,
                cv.FC %>% slice_min(cve) %>% pull(lambda.time),
                cv.CFC %>% slice_min(cve) %>% pull(lambda.time),
                cv.UC %>% slice_min(cve) %>% pull(lambda.time),
                cv.DR %>% slice_min(cve) %>% pull(lambda.time),
                cv.GCV %>% slice_min(cve) %>% pull(lambda.time),
                cv.FC %>% slice_min(cve) %>% pull(lambda.var),
                cv.CFC %>% slice_min(cve) %>% pull(lambda.var),
                cv.UC %>% slice_min(cve) %>% pull(lambda.var),
                cv.DR %>% slice_min(cve) %>% pull(lambda.var),
                cv.GCV %>% slice_min(cve) %>% pull(lambda.var))
            })
colnames(raw) <- rep(c("FC", "CFC", "UC", "DR", "GCV"), 5)
negunpenllkd.train.2000 <- raw[,1:5]
negunpenllkd.val.2000 <- raw[,6:10]
IMSE.2000 <- raw[,11:15]
lambda.time.2000 <- raw[,16:20]
lambda.var.2000 <- raw[,21:25]
colMeans(negunpenllkd.train.2000)
apply(negunpenllkd.train.2000, 2, sd)
colMeans(negunpenllkd.val.2000)
apply(negunpenllkd.val.2000, 2, sd)
colMeans(IMSE.2000)
apply(IMSE.2000, 2, sd)
apply(lambda.time.2000, 2, table)
apply(lambda.var.2000, 2, table)
save(negunpenllkd.train.2000, negunpenllkd.val.2000, IMSE.2000, lambda.time.2000, lambda.var.2000,
     file=paste0("CV_5methods_nrep=100_n=2000_4fold_",Sys.Date(), ".RData"))

colnames(raw) <- rep(c("FC", "CFC", "UC", "DR", "GCV"), 5)
negunpenllkd.train.5000 <- raw[,1:5]
negunpenllkd.val.5000 <- raw[,6:10]
IMSE.5000 <- raw[,11:15]
lambda.time.5000 <- raw[,16:20]
lambda.var.5000 <- raw[,21:25]
colMeans(negunpenllkd.train.5000)
apply(negunpenllkd.train.5000, 2, sd)
colMeans(negunpenllkd.val.5000)
apply(negunpenllkd.val.5000, 2, sd)
colMeans(IMSE.5000)
apply(IMSE.2000, 2, sd)
apply(lambda.time.5000, 2, table)
apply(lambda.var.5000, 2, table)
save(negunpenllkd.train.5000, negunpenllkd.val.5000, IMSE.5000, lambda.time.5000, lambda.var.5000,
     file=paste0("CV_5methods_nrep=100_n=5000_4fold_",Sys.Date(), ".RData"))

colnames(raw) <- rep(c("FC", "CFC", "UC", "DR", "GCV"), 5)
negunpenllkd.train.4000 <- raw[,1:5]
negunpenllkd.val.4000 <- raw[,6:10]
IMSE.4000 <- raw[,11:15]
lambda.time.4000 <- raw[,16:20]
lambda.var.4000 <- raw[,21:25]
colMeans(negunpenllkd.train.4000)
apply(negunpenllkd.train.4000, 2, sd)
colMeans(negunpenllkd.val.4000)
apply(negunpenllkd.val.4000, 2, sd)
colMeans(IMSE.4000)
apply(IMSE.4000, 2, sd)
apply(lambda.time.4000, 2, table)
apply(lambda.var.4000, 2, table)
save(negunpenllkd.train.4000, negunpenllkd.val.4000, IMSE.4000, lambda.time.4000, lambda.var.4000,
     file=paste0("CV_5methods_nrep=100_n=4000_4fold_",Sys.Date(), ".RData"))

colnames(raw) <- rep(c("FC", "CFC", "UC", "DR", "GCV"), 5)
negunpenllkd.train.3000 <- raw[,1:5]
negunpenllkd.val.3000 <- raw[,6:10]
IMSE.3000 <- raw[,11:15]
lambda.time.3000 <- raw[,16:20]
lambda.var.3000 <- raw[,21:25]
colMeans(negunpenllkd.train.3000)
apply(negunpenllkd.train.3000, 2, sd)
colMeans(negunpenllkd.val.3000)
apply(negunpenllkd.val.3000, 2, sd)
colMeans(IMSE.3000)
apply(IMSE.3000, 2, sd)
apply(lambda.time.3000, 2, table)
apply(lambda.var.3000, 2, table)
save(negunpenllkd.train.3000, negunpenllkd.val.3000, IMSE.3000, lambda.time.3000, lambda.var.3000,
     file=paste0("CV_5methods_nrep=100_n=3000_4fold_",Sys.Date(), ".RData"))