surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti,
                     nbs.time=6, nbs.var=8, degree.time=2, degree.var=3,
                     pen=F, lambda.time=NULL, lambda.var=NULL) {
  data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
  strata.withevent <-
    data %>% group_by(!!sym(strata)) %>%
    summarise(event.ct=sum(!!sym(event))) %>%
    filter(event.ct!=0) %>% pull(strata)
  data <- data %>% filter(!!sym(strata) %in% strata.withevent)
  knots.var <-
    quantile(data %>% pull(var),
             (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
  bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                           intercept=T, knots=knots.var) %>% as.matrix
  knots.time <- 
    quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
             (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
  t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
    group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
  bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                            knots=knots.time,
                            Boundary.knots=data %>% pull(time) %>% range) %>%
    as.matrix
  if (pen) {
    if (is.null(lambda.time) | is.null(lambda.var))
      stop("lambda.time and lambda.var should NOT be NULL!")
    fit <- surtiver_fixtra_bresties_fit_diffpen(
      data %>% pull(event), data %>% pull(time),
      sapply(split(data[,strata], data[,strata]), length),
      data %>% select(all_of(Z.tv)) %>% as.matrix,
      bases.var, bases.time, 
      matrix(0, length(Z.tv), nbs.time*nbs.var),
      data %>% select(all_of(Z.ti)) %>% as.matrix,
      matrix(0, length(Z.ti)),
      diff(diag(nbs.var)), diff(diag(nbs.time)), lambda.var, lambda.time)
  } else {
    fit <-
      surtiver_fixtra_bresties_fit(
        data %>% pull(event), data %>% pull(time),
        sapply(split(data[,strata], data[,strata]), length),
        data %>% select(all_of(Z.tv)) %>% as.matrix,
        bases.var, bases.time, 
        matrix(0, length(Z.tv), nbs.time*nbs.var),
        data %>% select(all_of(Z.ti)) %>% as.matrix,
        matrix(0, length(Z.ti)))
  }
  return(fit)
}


unpenllkd.surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti,
                               theta.time, beta.var, nbs.time=6, nbs.var=8,
                               degree.time=2, degree.var=3) {
  data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
  strata.withevent <-
    data %>% group_by(!!sym(strata)) %>%
    summarise(event.ct=sum(!!sym(event))) %>%
    filter(event.ct!=0) %>% pull(strata)
  data <- data %>% filter(!!sym(strata) %in% strata.withevent)
  knots.var <-
    quantile(data %>% pull(var),
             (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
  bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                           intercept=T, knots=knots.var) %>% as.matrix
  knots.time <- 
    quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
             (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
  t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
    group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
  bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                            knots=knots.time,
                            Boundary.knots=data %>% pull(time) %>% range) %>%
    as.matrix
  logplkd_fixtra_bresties_standalone(
    data %>% pull(event), data %>% pull(time),
    sapply(split(data[,strata], data[,strata]), length),
    data %>% select(all_of(Z.tv)) %>% as.matrix,
    bases.var, bases.time, theta.time,
    data %>% select(all_of(Z.ti)) %>% as.matrix,
    beta.var)
}

unpenllkd.surbiver.cv.uc <- function(data, event, time, var, strata, fold,
                                     Z.tv, Z.ti,
                                     theta.time.ls, beta.var.ls,
                                     nbs.time=6, nbs.var=8,
                                     degree.time=2, degree.var=3) {
  data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
  strata.withevent <-
    data %>% group_by(!!sym(strata)) %>%
    summarise(event.ct=sum(!!sym(event))) %>%
    filter(event.ct!=0) %>% pull(strata)
  data <- data %>% filter(!!sym(strata) %in% strata.withevent)
  knots.var <-
    quantile(data %>% pull(var),
             (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
  bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                           intercept=T, knots=knots.var) %>% as.matrix
  knots.time <- 
    quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
             (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
  t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
    group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
  bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                            knots=knots.time,
                            Boundary.knots=data %>% pull(time) %>% range) %>%
    as.matrix
  logplkd_fixtra_bresties_cv_UC(
    data %>% pull(event), data %>% pull(time),
    sapply(split(data[,strata], data[,strata]), length),
    data %>% pull(fold),
    data %>% select(all_of(Z.tv)) %>% as.matrix,
    bases.var, bases.time, theta.time.ls,
    data %>% select(all_of(Z.ti)) %>% as.matrix,
    beta.var.ls)
}

sumdrsq.surbiver.cv.dr <- function(data, event, time, var, strata, fold,
                                   Z.tv, Z.ti,
                                   theta.time.ls, beta.var.ls,
                                   nbs.time=6, nbs.var=8,
                                   degree.time=2, degree.var=3) {
  data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
  strata.withevent <-
    data %>% group_by(!!sym(strata)) %>%
    summarise(event.ct=sum(!!sym(event))) %>%
    filter(event.ct!=0) %>% pull(strata)
  data <- data %>% filter(!!sym(strata) %in% strata.withevent)
  knots.var <-
    quantile(data %>% pull(var),
             (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
  bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                           intercept=T, knots=knots.var) %>% as.matrix
  knots.time <- 
    quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
             (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
  t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
    group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
  bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                            knots=knots.time,
                            Boundary.knots=data %>% pull(time) %>% range) %>%
    as.matrix
  hazard <-
    surtiver_fixtra_bresties_fit(
      data %>% pull(event), data %>% pull(time),
      sapply(split(data[,strata], data[,strata]), length),
      data %>% select(all_of(Z.tv)) %>% as.matrix,
      bases.var, bases.time, 
      matrix(0, length(Z.tv), nbs.time*nbs.var),
      data %>% select(all_of(Z.ti)) %>% as.matrix,
      matrix(0, length(Z.ti)))$hazard
  sumDRsq_fixtra_bresties_cv_DR(
    data %>% pull(event), data %>% pull(time),
    sapply(split(data[,strata], data[,strata]), length),
    data %>% pull(fold),
    data %>% select(all_of(Z.tv)) %>% as.matrix,
    bases.var, bases.time, theta.time.ls,
    data %>% select(all_of(Z.ti)) %>% as.matrix,
    beta.var.ls, hazard)
}

cv.surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti, nbs.time=6,
                        nbs.var=8, degree.time=2, degree.var=3, nfold=5,
                        grid=expand_grid(lambda.time=10^(-(1:5)),
                                         lambda.var=10^(-(1:5))),
                        method="FC", seed=1, parallel=F) {
  if (method%in%c("FC", "CFC", "UC", "DR")) {
    set.seed(seed)
    data <- data %>% mutate(fold=sample.int(nfold, NROW(data), T))
    ngrid <- nrow(grid)
    if (parallel) {
      pkgs <- c("dplyr", "tibble", "magrittr", "tidyr", "Rcpp", "RcppArmadillo")
      noexport <- c("logplkd_fixtra_bresties_standalone",
                    "surtiver_fixtra_bresties_fit_diffpen",
                    "logplkd_fixtra_bresties_cv_UC")
      ncore <- parallel::detectCores()
      div.ngrid <- numbers::divisors(ngrid)
      ncore <- ifelse(ngrid<=ncore, ngrid,
                      max(div.ngrid[which(div.ngrid<=ncore)]))
      cl <- parallel::makeCluster(ncore)
      doParallel::registerDoParallel(cl)
      `%dopar%` <- foreach::`%dopar%`
      error.cv <- foreach::foreach(
        i=1:ngrid, .packages=pkgs, .combine=c, .noexport=noexport) %dopar% {
          Rcpp::sourceCpp("~/K/Users/kecc-wenbowu/Time_Varying_Effects/code/surtiver_tps.cpp")
          surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti,
                               nbs.time=6, nbs.var=8, degree.time=2, degree.var=3,
                               pen=F, lambda.time=NULL, lambda.var=NULL) {
            data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
            strata.withevent <-
              data %>% group_by(!!sym(strata)) %>%
              summarise(event.ct=sum(!!sym(event))) %>%
              filter(event.ct!=0) %>% pull(strata)
            data <- data %>% filter(!!sym(strata) %in% strata.withevent)
            knots.var <-
              quantile(data %>% pull(var),
                       (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
            bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                                     intercept=T, knots=knots.var) %>% as.matrix
            knots.time <- 
              quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
                       (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
            t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
              group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
            bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                                      knots=knots.time,
                                      Boundary.knots=data %>% pull(time) %>% range) %>%
              as.matrix
            if (pen) {
              if (is.null(lambda.time) | is.null(lambda.var))
                stop("lambda.time and lambda.var should NOT be NULL!")
              fit <- surtiver_fixtra_bresties_fit_diffpen(
                data %>% pull(event), data %>% pull(time),
                sapply(split(data[,strata], data[,strata]), length),
                data %>% select(all_of(Z.tv)) %>% as.matrix,
                bases.var, bases.time, 
                matrix(0, length(Z.tv), nbs.time*nbs.var),
                data %>% select(all_of(Z.ti)) %>% as.matrix,
                matrix(0, length(Z.ti)),
                diff(diag(nbs.var)), diff(diag(nbs.time)), lambda.var, lambda.time)
            } else {
              fit <-
                surtiver_fixtra_bresties_fit(
                  data %>% pull(event), data %>% pull(time),
                  sapply(split(data[,strata], data[,strata]), length),
                  data %>% select(all_of(Z.tv)) %>% as.matrix,
                  bases.var, bases.time, 
                  matrix(0, length(Z.tv), nbs.time*nbs.var),
                  data %>% select(all_of(Z.ti)) %>% as.matrix,
                  matrix(0, length(Z.ti)))
            }
            return(fit)
          }
          unpenllkd.surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti,
                                         theta.time, beta.var, nbs.time=6, nbs.var=8,
                                         degree.time=2, degree.var=3) {
            data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
            strata.withevent <-
              data %>% group_by(!!sym(strata)) %>%
              summarise(event.ct=sum(!!sym(event))) %>%
              filter(event.ct!=0) %>% pull(strata)
            data <- data %>% filter(!!sym(strata) %in% strata.withevent)
            knots.var <-
              quantile(data %>% pull(var),
                       (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
            bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                                     intercept=T, knots=knots.var) %>% as.matrix
            knots.time <- 
              quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
                       (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
            t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
              group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
            bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                                      knots=knots.time,
                                      Boundary.knots=data %>% pull(time) %>% range) %>%
              as.matrix
            logplkd_fixtra_bresties_standalone(
              data %>% pull(event), data %>% pull(time),
              sapply(split(data[,strata], data[,strata]), length),
              data %>% select(all_of(Z.tv)) %>% as.matrix,
              bases.var, bases.time, theta.time,
              data %>% select(all_of(Z.ti)) %>% as.matrix,
              beta.var)
          }
          unpenllkd.surbiver.cv.uc <- function(data, event, time, var, strata, fold,
                                               Z.tv, Z.ti,
                                               theta.time.ls, beta.var.ls,
                                               nbs.time=6, nbs.var=8,
                                               degree.time=2, degree.var=3) {
            data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
            strata.withevent <-
              data %>% group_by(!!sym(strata)) %>%
              summarise(event.ct=sum(!!sym(event))) %>%
              filter(event.ct!=0) %>% pull(strata)
            data <- data %>% filter(!!sym(strata) %in% strata.withevent)
            knots.var <-
              quantile(data %>% pull(var),
                       (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
            bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                                     intercept=T, knots=knots.var) %>% as.matrix
            knots.time <- 
              quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
                       (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
            t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
              group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
            bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                                      knots=knots.time,
                                      Boundary.knots=data %>% pull(time) %>% range) %>%
              as.matrix
            logplkd_fixtra_bresties_cv_UC(
              data %>% pull(event), data %>% pull(time),
              sapply(split(data[,strata], data[,strata]), length),
              data %>% pull(fold),
              data %>% select(all_of(Z.tv)) %>% as.matrix,
              bases.var, bases.time, theta.time.ls,
              data %>% select(all_of(Z.ti)) %>% as.matrix,
              beta.var.ls)
          }
          sumdrsq.surbiver.cv.dr <- function(data, event, time, var, strata, fold,
                                             Z.tv, Z.ti, theta.time.ls, beta.var.ls,
                                             nbs.time=6, nbs.var=8,
                                             degree.time=2, degree.var=3) {
            data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
            strata.withevent <-
              data %>% group_by(!!sym(strata)) %>%
              summarise(event.ct=sum(!!sym(event))) %>%
              filter(event.ct!=0) %>% pull(strata)
            data <- data %>% filter(!!sym(strata) %in% strata.withevent)
            knots.var <-
              quantile(data %>% pull(var),
                       (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
            bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                                     intercept=T, knots=knots.var) %>% as.matrix
            knots.time <- 
              quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
                       (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
            t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
              group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
            bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                                      knots=knots.time,
                                      Boundary.knots=data %>% pull(time) %>% range) %>%
              as.matrix
            hazard <-
              surtiver_fixtra_bresties_fit(
                data %>% pull(event), data %>% pull(time),
                sapply(split(data[,strata], data[,strata]), length),
                data %>% select(all_of(Z.tv)) %>% as.matrix,
                bases.var, bases.time, 
                matrix(0, length(Z.tv), nbs.time*nbs.var),
                data %>% select(all_of(Z.ti)) %>% as.matrix,
                matrix(0, length(Z.ti)))$hazard
            sumDRsq_fixtra_bresties_cv_DR(
              data %>% pull(event), data %>% pull(time),
              sapply(split(data[,strata], data[,strata]), length),
              data %>% pull(fold),
              data %>% select(all_of(Z.tv)) %>% as.matrix,
              bases.var, bases.time, theta.time.ls,
              data %>% select(all_of(Z.ti)) %>% as.matrix,
              beta.var.ls, hazard)
          }
          lambda.time <- grid %>% slice(n=i) %>% pull(lambda.time)
          lambda.var <- grid %>% slice(n=i) %>% pull(lambda.var)
          if (method%in%c("UC","DR")) {
            fits <-
              lapply(seq_len(nfold),
                     function(k, data, event, time, var, strata, Z.tv, Z.ti, nbs.time,
                              nbs.var, degree.time, degree.var, lambda.time, lambda.var) {
                       fit.k.c <-
                         surbiver(data %>% filter(fold!=k), event, time, var, strata,
                                  Z.tv, Z.ti, nbs.time, nbs.var, degree.time, degree.var,
                                  T, lambda.time, lambda.var)
                     }, data=data, event=event, time=time, var=var, strata=strata, Z.tv=Z.tv,
                     Z.ti=Z.ti, nbs.time=nbs.time, nbs.var=nbs.var, degree.time=degree.time,
                     degree.var=degree.var, lambda.time=lambda.time, lambda.var=lambda.var)
            theta.time.ls <- lapply(fits, function(fit) fit$ctrl.pts)
            beta.var.ls <- lapply(fits, function(fit) fit$tief)
            if (method=="UC") {
              unpenllkd.surbiver.cv.uc(data, event, time, var, strata, "fold",
                                       Z.tv, Z.ti, theta.time.ls, beta.var.ls,
                                       nbs.time, nbs.var, degree.time, degree.var)*(-2)
            } else {
              sumdrsq.surbiver.cv.dr(data, event, time, var, strata, "fold",
                                     Z.tv, Z.ti, theta.time.ls, beta.var.ls,
                                     nbs.time, nbs.var, degree.time, degree.var)
            }
          } else {        
            sapply(seq_len(nfold),
                   function(k, data, event, time, var, strata, Z.tv, Z.ti, nbs.time,
                            nbs.var, degree.time, degree.var, lambda.time, lambda.var) {
                     fit.k.c <-
                       surbiver(data %>% filter(fold!=k), event, time, var, strata, Z.tv, Z.ti,
                                nbs.time, nbs.var, degree.time, degree.var,
                                T, lambda.time, lambda.var)
                     if (method=="FC") { # fold constrained
                       unpenllkd.surbiver(data %>% filter(fold==k), event, time, var,
                                          strata, Z.tv, Z.ti,
                                          fit.k.c$ctrl.pts, fit.k.c$tief,
                                          nbs.time, nbs.var, degree.time, degree.var)
                     } else if (method=="CFC") { # complementary fold constrained
                       unpenllkd.surbiver(data, event, time, var,
                                          strata, Z.tv, Z.ti,
                                          fit.k.c$ctrl.pts, fit.k.c$tief,
                                          nbs.time, nbs.var, degree.time, degree.var) -
                         unpenllkd.surbiver(data %>% filter(fold!=k), event, time, var,
                                            strata, Z.tv, Z.ti,
                                            fit.k.c$ctrl.pts, fit.k.c$tief,
                                            nbs.time, nbs.var, degree.time, degree.var)
                     }
                   }, data=data, event=event, time=time, var=var, strata=strata, Z.tv=Z.tv,
                   Z.ti=Z.ti, nbs.time=nbs.time, nbs.var=nbs.var, degree.time=degree.time,
                   degree.var=degree.var, lambda.time=lambda.time, lambda.var=lambda.var) %>%
              sum %>% `*` (-2)
          }
        }
      parallel::stopCluster(cl)
    } else {
      error.cv <- apply(grid, 1, function(case) {
        names.case <- names(case)
        for (j in 1:length(case)) assign(names.case[j], unname(case[j]))
        if (method%in%c("UC","DR")) {
          fits <-
            lapply(seq_len(nfold),
                   function(k) {
                     fit.k.c <-
                       surbiver(data %>% filter(fold!=k), event, time, var, strata,
                                Z.tv, Z.ti, nbs.time, nbs.var, degree.time, degree.var,
                                T, lambda.time, lambda.var)
                   })
          theta.time.ls <- lapply(fits, function(fit) fit$ctrl.pts)
          beta.var.ls <- lapply(fits, function(fit) fit$tief)
          if (method=="UC") {
            unpenllkd.surbiver.cv.uc(data, event, time, var, strata, "fold",
                                     Z.tv, Z.ti, theta.time.ls, beta.var.ls,
                                     nbs.time, nbs.var, degree.time, degree.var)*(-2)
          } else { # deviance residual
            sumdrsq.surbiver.cv.dr(data, event, time, var, strata, "fold",
                                   Z.tv, Z.ti, theta.time.ls, beta.var.ls,
                                   nbs.time, nbs.var, degree.time, degree.var)
          }
        } else {
          sapply(seq_len(nfold),
                 function(k) {
                   fit.k.c <-
                     surbiver(data %>% filter(fold!=k), event, time, var, strata, Z.tv, Z.ti,
                              nbs.time, nbs.var, degree.time, degree.var,
                              T, lambda.time, lambda.var)
                   if (method=="FC") {
                     unpenllkd.surbiver(data %>% filter(fold==k), event, time, var,
                                        strata, Z.tv, Z.ti,
                                        fit.k.c$ctrl.pts, fit.k.c$tief,
                                        nbs.time, nbs.var, degree.time, degree.var)
                   } else if (method=="CFC") {
                     unpenllkd.surbiver(data, event, time, var,
                                        strata, Z.tv, Z.ti,
                                        fit.k.c$ctrl.pts, fit.k.c$tief,
                                        nbs.time, nbs.var, degree.time, degree.var) -
                       unpenllkd.surbiver(data %>% filter(fold!=k), event, time, var,
                                          strata, Z.tv, Z.ti,
                                          fit.k.c$ctrl.pts, fit.k.c$tief,
                                          nbs.time, nbs.var, degree.time, degree.var)
                   }
                 }) %>% sum %>% `*` (-2)
        }
      })
    }
  } else if (method=="GCV") { # GCV
    if (parallel) {
      ngrid <- nrow(grid)
      pkgs <- c("dplyr", "tibble", "magrittr", "tidyr", "Rcpp", "RcppArmadillo")
      noexport <- c("logplkd_fixtra_bresties_standalone",
                    "surtiver_fixtra_bresties_fit_diffpen")
      ncore <- parallel::detectCores()
      div.ngrid <- numbers::divisors(ngrid)
      ncore <- ifelse(ngrid<=ncore, ngrid,
                      max(div.ngrid[which(div.ngrid<=ncore)]))
      cl <- parallel::makeCluster(ncore)
      doParallel::registerDoParallel(cl)
      `%dopar%` <- foreach::`%dopar%`
      error.cv <- foreach::foreach(
        i=1:ngrid, .packages=pkgs, .combine=c, .noexport=noexport) %dopar% {
          Rcpp::sourceCpp("~/K/Users/kecc-wenbowu/Time_Varying_Effects/code/surtiver_tps.cpp")
          surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti,
                               nbs.time=6, nbs.var=8, degree.time=2, degree.var=3,
                               pen=F, lambda.time=NULL, lambda.var=NULL) {
            data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
            strata.withevent <-
              data %>% group_by(!!sym(strata)) %>%
              summarise(event.ct=sum(!!sym(event))) %>%
              filter(event.ct!=0) %>% pull(strata)
            data <- data %>% filter(!!sym(strata) %in% strata.withevent)
            knots.var <-
              quantile(data %>% pull(var),
                       (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
            bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                                     intercept=T, knots=knots.var) %>% as.matrix
            knots.time <- 
              quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
                       (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
            t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
              group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
            bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                                      knots=knots.time,
                                      Boundary.knots=data %>% pull(time) %>% range) %>%
              as.matrix
            if (pen) {
              if (is.null(lambda.time) | is.null(lambda.var))
                stop("lambda.time and lambda.var should NOT be NULL!")
              fit <- surtiver_fixtra_bresties_fit_diffpen(
                data %>% pull(event), data %>% pull(time),
                sapply(split(data[,strata], data[,strata]), length),
                data %>% select(all_of(Z.tv)) %>% as.matrix,
                bases.var, bases.time, 
                matrix(0, length(Z.tv), nbs.time*nbs.var),
                data %>% select(all_of(Z.ti)) %>% as.matrix,
                matrix(0, length(Z.ti)),
                diff(diag(nbs.var)), diff(diag(nbs.time)), lambda.var, lambda.time)
            } else {
              fit <-
                surtiver_fixtra_bresties_fit(
                  data %>% pull(event), data %>% pull(time),
                  sapply(split(data[,strata], data[,strata]), length),
                  data %>% select(all_of(Z.tv)) %>% as.matrix,
                  bases.var, bases.time, 
                  matrix(0, length(Z.tv), nbs.time*nbs.var),
                  data %>% select(all_of(Z.ti)) %>% as.matrix,
                  matrix(0, length(Z.ti)))
            }
            return(fit)
          }
          unpenllkd.surbiver <- function(data, event, time, var, strata, Z.tv, Z.ti,
                                         theta.time, beta.var, nbs.time=6, nbs.var=8,
                                         degree.time=2, degree.var=3) {
            data <- data %>% arrange(!!sym(strata), !!sym(time), desc(!!sym(event)))
            strata.withevent <-
              data %>% group_by(!!sym(strata)) %>%
              summarise(event.ct=sum(!!sym(event))) %>%
              filter(event.ct!=0) %>% pull(strata)
            data <- data %>% filter(!!sym(strata) %in% strata.withevent)
            knots.var <-
              quantile(data %>% pull(var),
                       (1:(nbs.var-degree.var-1))/(nbs.var-degree.var))
            bases.var <- splines::bs(data %>% pull(var), degree=degree.var,
                                     intercept=T, knots=knots.var) %>% as.matrix
            knots.time <- 
              quantile(data %>% filter(!!sym(event)!=0) %>% pull(time),
                       (1:(nbs.time-degree.time-1))/(nbs.time-degree.time))
            t.uniq.str <- data %>% filter(!!sym(event)!=0) %>%
              group_by(!!sym(strata)) %>% distinct(!!sym(time)) %>% pull(time)
            bases.time <- splines::bs(t.uniq.str, degree=degree.time, intercept=T,
                                      knots=knots.time,
                                      Boundary.knots=data %>% pull(time) %>% range) %>%
              as.matrix
            logplkd_fixtra_bresties_standalone(
              data %>% pull(event), data %>% pull(time),
              sapply(split(data[,strata], data[,strata]), length),
              data %>% select(all_of(Z.tv)) %>% as.matrix,
              bases.var, bases.time, theta.time,
              data %>% select(all_of(Z.ti)) %>% as.matrix,
              beta.var)
          }
          lambda.time <- grid %>% slice(n=i) %>% pull(lambda.time)
          lambda.var <- grid %>% slice(n=i) %>% pull(lambda.var)
          fit <- surbiver(data, event, time, var, strata, Z.tv, Z.ti,
                          nbs.time, nbs.var, degree.time, degree.var,
                          T, lambda.time, lambda.var)
          info.pen <- fit$info
          D.var <- diff(diag(nbs.var)); D.time <- diff(diag(nbs.time))
          A <- kronecker(kronecker(D.var, diag(nbs.time)),
                         lambda.var*diag(length(lambda.var)))
          B <- kronecker(kronecker(diag(nbs.var), D.time),
                         lambda.time*diag(length(lambda.time)))
          P <- (nrow(data)^1)*(t(A)%*%A + t(B)%*%B)
          info.pen[1:(nbs.var*nbs.time), 1:(nbs.var*nbs.time)] <-
            info.pen[1:(nbs.var*nbs.time), 1:(nbs.var*nbs.time)] - P
          -unpenllkd.surbiver(data, event, time, var, strata, Z.tv, Z.ti,
                              fit$ctrl.pts, fit$tief, nbs.time, nbs.var,
                              degree.time, degree.var)/nrow(data)/
            (1-(sum(diag(solve(fit$info)%*%info.pen))/nrow(data))^1)^2
        }
      parallel::stopCluster(cl)
    } else {
      error.cv <- apply(grid, 1, function(case) {
        names.case <- names(case)
        for (j in 1:length(case)) assign(names.case[j], unname(case[j]))
        fit <- surbiver(data, event, time, var, strata,
                        Z.tv, Z.ti, nbs.time, nbs.var, degree.time, degree.var,
                        T, lambda.time, lambda.var)
        info.pen <- fit$info
        D.var <- diff(diag(nbs.var)); D.time <- diff(diag(nbs.time))
        A <- kronecker(kronecker(D.var, diag(nbs.time)),
                       lambda.var*diag(length(lambda.var)))
        B <- kronecker(kronecker(diag(nbs.var), D.time),
                       lambda.time*diag(length(lambda.time)))
        P <- (nrow(data)^1)*(t(A)%*%A + t(B)%*%B)
        info.pen[1:(nbs.var*nbs.time), 1:(nbs.var*nbs.time)] <-
          info.pen[1:(nbs.var*nbs.time), 1:(nbs.var*nbs.time)] - P
        -unpenllkd.surbiver(data, event, time, var, strata, Z.tv, Z.ti,
                            fit$ctrl.pts, fit$tief, nbs.time, nbs.var,
                            degree.time, degree.var)/nrow(data)/
          (1-sum(diag(solve(fit$info)%*%info.pen))/nrow(data))^2
      })
    }
  }
  return(grid %>% mutate(cve=error.cv))
}